#include <curl/curl.h>
#include <storage/file.h>
#include <storage/directory.h>
#include <storage/filereference.h>
#include <util/messenger.h>
#include <util/datetime.h>

#include "job.h"
#include "server.h"
#include "remotenode.h"
#include "messages.h"

/* Include: extern int ftpparse(struct ftpparse *,const char *,int); */
extern "C" {
#include "ftpparse.h"
};

#define DEBUG( arg... ) do { if( m_pcServer->GetSettings()->GetDebugMode() ) printf( arg ); } while(0)

/** \brief Job Constructor.
 * Initializes the Job storing all the data related to the job.
 */
Job::Job( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nType, int nID, Job* pcParent )
{
	m_pcServer = pcServer;
	DEBUG( "Job %i created; parent %i\n", nID, pcParent==NULL?-1:pcParent->m_nID );	
	
	m_zLocalPath = zLocalPath;
	m_zRemotePath = zRemotePath;
	m_nType = nType;
	m_nStatus = JOB_QUEUED;
	m_nID = nID;
	m_pcParent = pcParent;
	m_psHandle = NULL;
	
	m_nTotalSize = -1;		/* -1 indicates 'not set yet' */
	m_nTotalDownloaded = -1;
	m_nTotalUploaded = -1;
}


/** \brief Job Default constructor.
 */
Job::~Job()
{
}

bool Job::IsSchedulable()
{
	if( m_pcParent == NULL ) return( true );
	else return( false );
}

bool Job::GetJobInfo( Message* pcMsg )
{
	pcMsg->AddInt32( "jobID", m_nID );
	pcMsg->AddInt32( "status", m_nStatus );
	
	if( m_nStatus == JOB_DELETING ) {
		pcMsg->AddBool( "finished", true );
	}
	
	return( true );
}

/** \brief Set the necessary options in the curl handle.
 * Here we set the basic options like callbacks. We also associate ourself with the CurlHandle_s by setting m_nStatus.
 */
CURLcode Job::AttachToHandle( CurlHandle_s* psHandle )
{
	m_psHandle = psHandle;
	
	CURL* pHandle = psHandle->m_pHandle;
	
	/* Remove easy handle from the multi handle, then re-add it at the end. */
	/* This is a workaround for (what seems to be) a libcurl bug. */
	/* Without doing this, sometimes curl doesn't notice the changes and doesn't run the transfer. */
	curl_multi_remove_handle( psHandle->m_pMultiHandle, psHandle->m_pHandle );
	
	/* Call curl_easy_setopt() to set url, callbacks, etc. */
	/* We set some general options here; specifics like resume position etc will be set by subclasses. */

	/* Set callback functions */
	curl_easy_setopt( pHandle, CURLOPT_WRITEFUNCTION, Job::WriteCallback );
	curl_easy_setopt( pHandle, CURLOPT_WRITEDATA, this );
	curl_easy_setopt( pHandle, CURLOPT_READFUNCTION, Job::ReadCallback );
	curl_easy_setopt( pHandle, CURLOPT_READDATA, this );
//	curl_easy_setopt( pHandle, CURLOPT_SEEKFUNCTION, Job::SeekCallback );	/* Not available in older versions of libcurl */
//	curl_easy_setopt( pHandle, CURLOPT_SEEKDATA, this );
	curl_easy_setopt( pHandle, CURLOPT_PROGRESSFUNCTION, Job::ProgressBarCallback );
	curl_easy_setopt( pHandle, CURLOPT_PROGRESSDATA, this );
	curl_easy_setopt( pHandle, CURLOPT_NOPROGRESS, 0 );
	
	psHandle->m_nStatus = m_nID;
	m_nStatus = JOB_RUNNING;

	curl_multi_add_handle( psHandle->m_pMultiHandle, psHandle->m_pHandle );	
	return( CURLE_OK );
}


/** \brief Remove any settings from the curl handle set in AttachToHandle().
 * \note This will be called when the transfer finishes, but it may also be called for instance
 *  if the user pauses the transfer and we need to use the curl handle for something else.
 */
CURLcode Job::RemoveFromHandle( CurlHandle_s* psHandle )
{
	DEBUG( "Job::RemoveFromHandle() job %i from curl handle %x\n", m_nID, psHandle->m_pHandle );
	CURL* pHandle = psHandle->m_pHandle;
	curl_easy_setopt( pHandle, CURLOPT_URL, "" );
	curl_easy_setopt( pHandle, CURLOPT_WRITEDATA, NULL );
	curl_easy_setopt( pHandle, CURLOPT_READDATA, NULL );
//	curl_easy_setopt( pHandle, CURLOPT_SEEKDATA, NULL );
	curl_easy_setopt( pHandle, CURLOPT_PROGRESSDATA, NULL );
	
	m_psHandle = NULL;
	psHandle->m_nStatus = HANDLE_IDLE;
	
	return( CURLE_OK );
}

void Job::TransferFinished( CurlHandle_s* psHandle, long nCode, Job* pcJob )
{
	DEBUG( "Transfer for job %i finished; result code %li\n:", m_nID, nCode );

	/* Non-recursive jobs will want to:
		- detach self from curl handle
		- flush data to disk
		- remove self from server lists
		- delete self
	  We implement this (except for flushing) here for the convenience of DownloadJob, etc.
	*/
	
	RemoveFromHandle( psHandle );
	m_nStatus = JOB_DELETING;
	m_pcServer->RemoveJob( this );

	/* If this job was created recursively, then pass the call on to our parent. */
	if( m_pcParent != NULL )
	{
		m_pcParent->TransferFinished( psHandle, nCode, this );
		return;
	}
}

void Job::Pause()
{
	if( m_psHandle ) RemoveFromHandle( m_psHandle );
	m_nStatus = JOB_PAUSED;
}

void Job::Resume()
{
	m_nStatus = JOB_QUEUED;
}

void Job::Cancel()
{
	if( m_psHandle ) RemoveFromHandle( m_psHandle );
	m_nStatus = JOB_DELETING;
	m_pcServer->RemoveJob( this );
}

/** \brief Read Callback
 * The libcurl read callback.
 * Can't pass a virtual function to libcurl so we use these static wrappers.
 *
 * \param pBuf The location to store input data.
 * \param nSize The amount of data to be read.
 * \param nMult The size in bytes of each data unit.
 * \param pCookie This contains the Job* that we want to call the Read method on.
 *
 */
size_t Job::ReadCallback( void* pBuf, size_t nSize, size_t nMult, void* pCookie )
{
	Job* pcTransfer = (Job*)pCookie;
	return( pcTransfer->Read( pBuf, nSize * nMult ) );
}


/** \brief Write Callback
 * The libcurl write callback.
 * Can't pass a virtual function to libcurl so we use these static wrappers.
 *
 * \todo Verify descriptions of callback parameters with libcurl documentation.
 *
 * \param pBuf The data to be stored.
 * \param nSize The size of the data being passed to the Job.
 * \param nMult The size of the data type being passed.
 * \param pCookie This contains the Job* that we want to call the Write method on.
 *
 */
size_t Job::WriteCallback( void* pBuf, size_t nSize, size_t nMult, void* pCookie )
{
	Job* pcTransfer = (Job*)pCookie;
	return( pcTransfer->Write( pBuf, nSize * nMult ) );
}

/** \brief Seek Callback
 * The libcurl seek callback.
 * Can't pass a virtual function to libcurl so we use these static wrappers.
 *
 * \param pCookie This contains the Job* that we want to call the Seek method on.
 * \param nOffset How many bytes to offset into the file.
 * \param nOrigin The origin from which to start the offset.
 */
int Job::SeekCallback( void* pCookie, curl_off_t nOffset, int nOrigin )
{
	Job* pcTransfer = (Job*)pCookie;
	return( pcTransfer->Seek( nOffset, nOrigin ) );
}

/** \brief Progress Bar Callback
 * The libcurl progress bar. callback.
 * Can't pass a virtual function to libcurl so we use these static wrappers.
 *
 * \param pCookie The pointer to the Job object to send the ProgreesBar update to.
 * \param fDownTotal Total data downloaded to be downloaded
 * \param fDownNow Total data downloaded so far.
 * \param fUpTotal Total data to be uploaded.
 * \param fUpNow Total data uploaded so far.
 */
int Job::ProgressBarCallback( void* pCookie, double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
	Job* pcTransfer = (Job*)pCookie;
	return( pcTransfer->ProgressBar( fDownTotal, fDownNow, fUpTotal, fUpNow ) );
}

/***** DownloadJob implementation ******/

/** \brief DownloadJob Constructor. 
 * Sets the initial download file location to null.
 */
DownloadJob::DownloadJob( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nID, Job* pcParent )
	: Job( pcServer, zLocalPath, zRemotePath, JOB_DOWNLOAD, nID, pcParent )
{
	m_pcFile = NULL;	/* This is only created when AttachToHandle() is first called. */
}

/** \brief DownloadJob Deconstructor.
 * This frees up the file pointer m_pcFile if necessary.
 */
DownloadJob::~DownloadJob()
{
	if( m_pcFile )
		delete( m_pcFile );
}

/** \brief Setup Function for Curl Transfer.
 * Configures the given CURL handle to perform this transfer, such as setting the url.
 */
CURLcode DownloadJob::AttachToHandle( CurlHandle_s* psHandle )
{
	String zServer = m_pcServer->GetServerAddress(); 	/* Address of the form "ftp://user:pwd@host:port" (no path) */
	if( m_pcFile == NULL ) {
		m_pcFile = new File( m_zLocalPath, O_WRONLY | O_CREAT | O_TRUNC );
	}
	
	Job::AttachToHandle( psHandle );
	

	curl_off_t nOffset = m_pcFile->Seek( 0, SEEK_CUR );	/* Get current position in file, for if we are resuming */
	
	/* Set the url, set to download */
	String zURL = m_pcServer->GetServerAddress();
	zURL += "/";
	zURL += m_zRemotePath;
	DEBUG( "DownloadJob: setting url to %s\n", zURL.c_str() );
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_URL, zURL.c_str() );
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_UPLOAD, 0 );
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_RESUME_FROM_LARGE, nOffset );
	
	return( CURLE_OK );
}

CURLcode DownloadJob::RemoveFromHandle( CurlHandle_s* psHandle )
{
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_RESUME_FROM_LARGE, (curl_off_t)0 );
	return( Job::RemoveFromHandle( psHandle ) );
}

void DownloadJob::TransferFinished( CurlHandle_s* psHandle, long nCode, Job* pcJob )
{
	if( m_pcFile ) m_pcFile->Flush();
	
	/* Should always be the last thing before returning */
	Job::TransferFinished( psHandle, nCode, pcJob );
}

bool DownloadJob::GetJobInfo( Message* pcMsg )
{
	pcMsg->AddString( "destPath", m_zLocalPath );
	pcMsg->AddString( "sourcePath", m_zRemotePath );
	pcMsg->AddInt32( "totalTransferred", m_nTotalDownloaded );
	pcMsg->AddInt32( "totalSize", m_nTotalSize );

	switch( m_nStatus ) {
		case JOB_RUNNING: pcMsg->AddString( "jobStatus", "Downloading" ); break;
		case JOB_PAUSED: pcMsg->AddString( "jobStatus", "Paused (download)" ); break;
		case JOB_QUEUED: pcMsg->AddString( "jobStatus", "Queued (download)" ); break;
		case JOB_DELETING: pcMsg->AddString( "jobStatus", "Finished (download)" ); break;
		default: pcMsg->AddString( "jobStatus", "Unknown! (download)" ); break;
	}
	return( Job::GetJobInfo( pcMsg ) );
}

void DownloadJob::Pause()
{
	if( m_pcFile ) m_pcFile->Flush();
	Job::Pause();
}

void DownloadJob::Cancel()
{
	if( m_pcFile ) m_pcFile->Flush();
	Job::Cancel();
}

/** \brief Write to file.
 * This performs the actual writing of data to the file.
 *
 * \param pBuf The data to write to the file.
 * \param nSize The size of the data being written.
 *
 * \return The amount of data actually written.
 */
size_t DownloadJob::Write( void* pBuf, size_t nSize )
{
	size_t nResult = m_pcFile->Write( pBuf, nSize );
//	m_pcFile->Flush();
	return( nResult );
}

/** \brief Performs a seek operation on the file.
 * Seek into a particular position in the file so curl can begin operating from there.
 *
 * \param nOffset The offset into the file to start operating at.
 * \param nOrigin The origin of the offset.
 *
 * \todo Make sure that File::Seek() returns the same as libcurl expects.
 */
int DownloadJob::Seek( curl_off_t nOffset, int nOrigin )
{
	return( m_pcFile->Seek( nOffset, nOrigin ) );

}

/** \brief Updates the progress bar.
 * This just outputs progress information to the terminal.
 *
 * \param fDownTotal Total bytes to be downloaded.
 * \param fDownNow Total bytes downloaded so far.
 * \param fUpTotal Total bytes to upload.
 * \param fUpNow Total bytes uploaded at the moment.
 */
int DownloadJob::ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
	SetDownloadedTotal( (size_t)fDownNow );
	SetTotalSize( (size_t)fDownTotal );
	m_pcServer->JobUpdated( this );
//	DEBUG( "Progress for download %s: %.1f%%\n", m_zRemotePath.c_str(), fDownNow/fDownTotal * 100 );
	return( 0 );
}

/****** UploadJob implementation ********/

/** \brief UploadJob Constructor. 
 * Sets the initial upload file location to null.
 */
UploadJob::UploadJob( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nID, Job* pcParent )
	: Job( pcServer, zLocalPath, zRemotePath, JOB_UPLOAD, nID, pcParent )
{
	m_pcFile = NULL;	/* This is only created when AttachToHandle() is first called. */
}

/** \brief UploadJob Deconstructor.
 * Frees the pointer to the upload file.
 */
UploadJob::~UploadJob()
{
	if( m_pcFile )
		delete( m_pcFile );
}

/** \brief Setup Handle for Curl Transfer.
 * This does the setup for an upload curl handler.
 */
CURLcode UploadJob::AttachToHandle( CurlHandle_s* psHandle )
{
	String zServer = m_pcServer->GetServerAddress(); 	/* Address of the form "ftp://user:pwd@host:port" (no path) */
	if( m_pcFile == NULL ) {
		m_pcFile = new File( m_zLocalPath, O_RDONLY );
	}
	
	Job::AttachToHandle( psHandle );
	

#if 1
	/* I don't think resuming uploads is this easy with libcurl :( */
	curl_off_t nOffset = m_pcFile->Seek( 0, SEEK_CUR );	/* Get current position in file, for if we are resuming */
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_RESUME_FROM_LARGE, nOffset );
#else
	m_pcFile->Seek( 0, SEEK_SET );
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_RESUME_FROM_LARGE, (curl_off_t)0 );
#endif

	curl_off_t nSize = m_pcFile->GetSize();
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_INFILESIZE_LARGE, nSize );
	
	/* Set the url, set to upload */
	String zURL = m_pcServer->GetServerAddress();
	zURL += "/";
	zURL += m_zRemotePath;
	DEBUG( "UploadJob: setting url to %s\n", zURL.c_str() );
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_URL, zURL.c_str() );
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_UPLOAD, 1 );
	
	return( CURLE_OK );
}

CURLcode UploadJob::RemoveFromHandle( CurlHandle_s* psHandle )
{
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_RESUME_FROM_LARGE, (curl_off_t)0 );
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_INFILESIZE_LARGE, 0 );
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_UPLOAD, 0 );
	
	return( Job::RemoveFromHandle( psHandle ) );
}

/** \brief Read from file.
 * This performs the read from the file and returns the size read.
 *
 * \param pBuf A pointer to a location that the data can be read into.
 * \param nSize The amount of data to be read.
 */
size_t UploadJob::Read( void* pBuf, size_t nSize )
{
	return( m_pcFile->Read( pBuf, nSize ) );
}

/** \brief Seek into the upload file.
 * Seek into a particular position in the file so curl can begin operating from there.
 *
 * \param nOffset The offset into the file to start operating at.
 * \param nOrigin The origin of the offset.
 *
 * \todo Make sure that File::Seek() returns the same as libcurl expects.
 */
int UploadJob::Seek( curl_off_t nOffset, int nOrigin )
{
	return( m_pcFile->Seek( nOffset, nOrigin ) );
	
}

/** \brief Progress Bar Update.
 * This just outputs progress information to the terminal.
 *
 * \param fDownTotal Total bytes to be downloaded.
 * \param fDownNow Total bytes downloaded so far.
 * \param fUpTotal Total bytes to upload.
 * \param fUpNow Total bytes uploaded at the moment.
 */
int UploadJob::ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
	SetUploadedTotal( (size_t)fUpNow );
	SetTotalSize( (size_t)fUpTotal );
	m_pcServer->JobUpdated( this );
//	DEBUG( "Progress for upload %s: %.1f%%\n", m_zRemotePath.c_str(), (float)(100 * fUpNow/fUpTotal) );
	return( 0 );
}

bool UploadJob::GetJobInfo( Message* pcMsg )
{
	pcMsg->AddString( "sourcePath", m_zLocalPath );
	pcMsg->AddString( "destPath", m_zRemotePath );
	pcMsg->AddInt32( "totalTransferred", m_nTotalDownloaded );
	pcMsg->AddInt32( "totalSize", m_nTotalSize );

	switch( m_nStatus ) {
		case JOB_RUNNING: pcMsg->AddString( "jobStatus", "Uploading" ); break;
		case JOB_PAUSED: pcMsg->AddString( "jobStatus", "Paused (upload)" ); break;
		case JOB_QUEUED: pcMsg->AddString( "jobStatus", "Queued (upload)" ); break;
		case JOB_DELETING: pcMsg->AddString( "jobStatus", "Finished (upload)" ); break;
		default: pcMsg->AddString( "jobStatus", "Unknown! (upload)" ); break;
	}
	
	return( Job::GetJobInfo( pcMsg ) );
}


/****** DirListJob implementation *******/

/** \brief Default constructor for DirListJob.
 * This initializes the target of the directory listing and makes note that
 * this is starting the listing.
 */
DirListJob::DirListJob( Server* pcServer, const String& zRemotePath, int nID, Job* pcParent )
	: Job( pcServer, "", zRemotePath, JOB_DIRLIST, nID, pcParent )
{
	DEBUG( "Job %i is a DirListNode for %s\n", nID, zRemotePath.c_str() );
	m_bInitial = true;
}

/** \brief Default deconstructor for DirListJob.
 */
DirListJob::~DirListJob()
{
}

/** \brief Setup the Curl handle for the directory listing.
 */
CURLcode DirListJob::AttachToHandle( CurlHandle_s* psHandle )
{
	String zServer = m_pcServer->GetServerAddress(); 	/* Address of the form "ftp://user:pwd@host:port" (no path) */
	
	Job::AttachToHandle( psHandle );
	
	String zURL = zServer;
	zURL += "/";
	zURL += m_zRemotePath;
	zURL += "/";			/* Make it clear to libcurl that we want a dir listing */
	

	/* Set the url, set to upload */
	DEBUG( "DirListJob:AttachToHandle(): setting url to %s\n", zURL.c_str() );
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_URL, zURL.c_str() );
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_UPLOAD, 0 );
//	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_RESUME_FROM_LARGE, (curl_off_t)0 );
	
	return( CURLE_OK );
}

/** \brief Find the first occurrence of nChar in the first nLength bytes of pzStr. 
 * If not found, points to the first char *after* the end of the string.
 */
const char* strnchr( const char* pzStr, size_t nLength, int nChar )
{
	size_t nPos = 0;
	while( nPos < nLength && pzStr[nPos] != nChar && pzStr[nPos] != 0 ) ++nPos;
	return( pzStr + nPos );
}

/** \brief Parse (part of) a dirlisting into a list of filesystem nodes.
 * This functionality is used by RecursiveDownloadJob also.
 */
size_t DirListJob::ParseDirListData( const char* pzBuf, size_t nSize, std::vector< RemoteNode >* pacList )
{
	struct ftpparse sData;
	memset( &sData, 0, sizeof( sData ) );
	
	size_t nTotalRead = 0; // Total bytes read.
	String zLine = m_zLastLineFragment; // Copy the last fragment read if we didn't read a complete line at the end last time.
	size_t nLineLength = m_zLastLineFragment.size(); // Initialize the current line length.
	const char* pzLine = pzBuf;
	
	// Read in the next line up to and excluding the '\n'
	nTotalRead = strnchr( pzLine, nSize, '\n' ) - pzLine;
	nLineLength += nTotalRead;
	
	// Append the new text to the end of the line.
	zLine += String( pzLine, nTotalRead );
	
	/* Didn't find a \n; the line is incomplete so save it for later */
	if( nTotalRead == nSize )
	{
		m_zLastLineFragment = zLine;
		return( nSize );
	}
	
	do
	{
//		DEBUG( "pzBuf: %x  pzLine: %x  length: %i     pzLine - pzBuf: %i\n", pzBuf, pzLine, nLineLength, pzLine - pzBuf );
//		DEBUG( "Line: '%s' length: %i\n", zLine.c_str(), nLineLength );

		// Attempt to parse the current line.
		int nResult;
		nResult = ftpparse( &sData, zLine.c_str(), nLineLength );
//		DEBUG( "Parse result: %i\n", nResult );

		if( nResult == 1 )	/* 1 indicates success */
		{
			// Create a new RemoteNode for the directory entry.
			RemoteNode cNode;
			cNode.m_zName = String( sData.name, sData.namelen );
			cNode.m_bIsDir = sData.flagtrycwd;
			cNode.m_nSize = sData.size;
			cNode.m_cTimestamp = DateTime( sData.mtime );
			cNode.m_zPath.Format( "%s/%s", m_zRemotePath.c_str(), cNode.m_zName.c_str() );
			/* ftpparse doesn't provide permissions! Guess we'll do without for now. */
			cNode.m_nPermissions = 0777;
//			DEBUG( "Parsed: name '%s', path '%s', size %i, %s, %s\n", cNode.m_zName.c_str(), cNode.m_zPath.c_str(), cNode.m_nSize, (cNode.m_bIsDir?"dir":"file"), cNode.m_cTimestamp.GetDate().c_str() );
			pacList->push_back( cNode );
		}
		else
		{
			String zTmp = String( pzLine, nLineLength );	/* Just for the next printf */
			printf( "ParseDirListing: ftpparse() failed on '%s' (%i bytes)\n", zTmp.c_str(), nLineLength );
		}

		nTotalRead++;	/* Skip the \n */
		pzLine = pzBuf + nTotalRead;
		nLineLength = strnchr( pzLine, nSize - nTotalRead, '\n' ) - pzLine;
		nTotalRead += nLineLength;
		zLine = String( pzLine, nLineLength );
		
		
		/* Save anything after the last \n (this will be empty if the pzBuf ends with \n) */
		if( nTotalRead >= nSize )
		{
			m_zLastLineFragment = zLine;
			break;
		}
	} while( 1 );
	
//	DEBUG( " nTotalRead: %i   nSize: %i   zLastLineFragment: '%s'\n", nTotalRead, nSize, m_zLastLineFragment.c_str() );	
	return( nSize );
}

/** \brief Write the data into the data buffer.
 * This "writes" the data coming from the remote server into the RemoteNode
 * objects. These are then passed to RemoteIconView to display to the user.
 *
 * \warning The RemoteNode vector that is sent is assumed to be deleted by the
 * RemoteIconView that calls for the update when it is done processing.
 *
 * \bug Generated nodes do not have proper permissions due to a deficiency in
 * how ftpparse operates.
 *
 * \todo The dirlist parsing might be better in Server, but we need to do it here because we need to save any fragments of the last line.
 *
 * \todo Fix up initial, final flags that are sent to the RemoteIconView. Maybe remove them. Currently the 'final' flag isn't being sent.
 */
size_t DirListJob::Write( void* pBuf, size_t nSize )
{
	/* If ftp entry path hasn't been determined yet, do it now. */
	/* We do this here, since we want to provide the entry path to the GUI at the earliest opportunity */
	if( m_pcServer->m_zEntryPath == "" )
	{
		const char* pzEntryPath = NULL;
		curl_easy_getinfo( m_psHandle->m_pHandle, CURLINFO_FTP_ENTRY_PATH, &pzEntryPath );
		if( pzEntryPath == NULL ) {
			printf( "DirListJob: FTP_ENTRY_PATH returned NULL!\n" );
		} else {
			m_pcServer->_SetEntryPath( pzEntryPath );
			DEBUG( "DirListJob: Got entry path %s\n", pzEntryPath );
		}
	}

	char* pzBuf = (char*)pBuf;
	String zTmp( pzBuf, nSize );
//	DEBUG( "DirListJob::Write( %s, %i bytes, %sinitial ):\n%s\n------ end listing data ------\n", m_zRemotePath.c_str(), (int)nSize, m_bInitial?"":"not ", zTmp.c_str() );

	std::vector< RemoteNode >* pacList = new std::vector< RemoteNode >;	/* Will be deleted by the RemoteIconView when it's finished processing */
	size_t nResult = ParseDirListData( pzBuf, nSize, pacList );

	if( pacList->size() == 0 )
	{
		/* We didn't sucessfully parse any entries! Clean up. */
		delete( pacList );
		return( nResult );
	}
	
	HandleDirListData( pacList );
	
	m_bInitial = false;
	
	return( nResult );
}

/** \brief Update the Progress bar for the job.
 * This just outputs progress information to the terminal.
 *
 * \param fDownTotal Total bytes to be downloaded.
 * \param fDownNow Total bytes downloaded so far.
 * \param fUpTotal Total bytes to upload.
 * \param fUpNow Total bytes uploaded at the moment.
 */
int DirListJob::ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
//	DEBUG( "Progress for dirlisting %s: %.1f out of %.1f\n", m_zRemotePath.c_str(), (float)fDownNow, (float)fDownTotal );
	return( 0 );
}

void DirListJob::HandleDirListData( std::vector< RemoteNode >* pacList )
{
	DEBUG( "Warning: DirListJob::HandleDirListData() called!\n" );
}

bool DirListJob::GetJobInfo( Message* pcMsg )
{
	return( false );
}

/*************** GUIDirListJob implementation ************************/

GUIDirListJob::GUIDirListJob( Server* pcServer, const String& zRemotePath, Handler* pcTarget, int nID, Job* pcParent )
	: DirListJob( pcServer, zRemotePath, nID, pcParent )
{
	m_pcTarget = pcTarget;
}

GUIDirListJob::~GUIDirListJob()
{
}

void GUIDirListJob::HandleDirListData( std::vector< RemoteNode >* pacList )
{

	Message cMsg( M_REMOTE_DIRLISTING );
	cMsg.AddString( "path", m_zRemotePath );
	cMsg.AddPointer( "list", pacList );

	if( m_bInitial )
		cMsg.AddBool( "initial", true );	/* TODO: fix up initial, final flags. Maybe remove them. */
//	cMsg.AddBool( "final", true );
	
	Messenger cMessenger( m_pcTarget );
	cMessenger.SendMessage( &cMsg );
}

void GUIDirListJob::TransferFinished( CurlHandle_s* psHandle, long nCode, Job* pcJob )
{
	/* TODO: notify GUI that listing has finished */
	DirListJob::TransferFinished( psHandle, nCode, pcJob );
}

/**** CommandJob implementation *******/

CommandJob::CommandJob( Server* pcServer, const std::vector<String>& azCommands, int nID, Job* pcParent )
 : Job( pcServer, "", "", JOB_COMMAND, nID, pcParent )
{
	DEBUG( "Job %i is a CommandJob doing %s\n", nID, azCommands.front().c_str() );
	m_azCommands = azCommands;
	m_psCommands = NULL;
}

CommandJob::~CommandJob()
{
}

CURLcode CommandJob::AttachToHandle( CurlHandle_s* psHandle )
{
	DEBUG( "CommandJob::AttachToHandle(): '%s'\n", m_azCommands.front().c_str() );
	Job::AttachToHandle( psHandle );

	String zURL = m_pcServer->GetServerAddress();
	zURL += "/";	/* Dummy path */
	
	/* Attach the commands to the handle */
	std::vector<String>::iterator i = m_azCommands.begin();
	for( ; i != m_azCommands.end(); i++ ) {
		m_psCommands = curl_slist_append( m_psCommands, i->c_str() );
	}

	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_QUOTE, m_psCommands );
	
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_URL, zURL.c_str() );
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_NOBODY, 1 );

	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_NOPROGRESS, 1 );	/* Don't want progress notifications */

	return( CURLE_OK );
}

CURLcode CommandJob::RemoveFromHandle( CurlHandle_s* psHandle )
{
	DEBUG( "CommandJob::RemoveFromHandle() job %i\n", m_nID );

	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_NOBODY, 0 );
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_QUOTE, NULL );
	curl_easy_setopt( psHandle->m_pHandle, CURLOPT_NOPROGRESS, 0 );
	
	curl_slist_free_all( m_psCommands );
	
	return( Job::RemoveFromHandle( psHandle ) );
}

bool CommandJob::GetJobInfo( Message* pcMsg )
{
	return( false );
}

/**** RecursiveDownloadJob implementation ****/

RecursiveDownloadJob::RecursiveDownloadJob( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nID, Job* pcParent )
	: DirListJob( pcServer, zRemotePath, nID, pcParent )
{
	m_nType = JOB_RECURSIVE;
	m_zLocalPath = zLocalPath;
	m_zRemotePath = zRemotePath;
	
	m_pIterator = m_acList.begin();
}

RecursiveDownloadJob::~RecursiveDownloadJob()
{
}

CURLcode RecursiveDownloadJob::AttachToHandle( CurlHandle_s* psHandle )
{
	/* Create the destination dir, if it doesn't already exist */
	/* TODO: This really belongs in an Initialize() function, not in AttachToHandle(),
	   since AttachToHandle() could potentially be called more than once if the transfer is paused then resumed.
	 */
	int nResult = mkdir( m_zLocalPath.c_str(), S_IRWXU );
	if( nResult != 0 && errno != EEXIST ) {
		/* TODO: what if a file of the same name exists but isn't a dir? */
		printf( "RecursiveDownloadJob: mkdir( %s ) failed! %i %s\n", m_zLocalPath.c_str(), errno, strerror(errno) );
		return( (CURLcode) -CURLE_FAILED_INIT );	/* Just need to return nonzero here - this seems a suitable code */
	}
	if( nResult == 0 ) DEBUG( "RecursiveDownloadJob: Created local dir %s\n", m_zLocalPath.c_str() );
	else DEBUG( "RecursiveDownloadJob: Local dir/file %s already exists!\n", m_zLocalPath.c_str() );

	return( DirListJob::AttachToHandle( psHandle ) );
}

void RecursiveDownloadJob::HandleDirListData( std::vector< RemoteNode >* pacList )
{
	m_acList.insert( m_acList.end(), pacList->begin(), pacList->end() );
	m_pIterator = m_acList.begin();
	delete( pacList );
}

void RecursiveDownloadJob::TransferFinished( CurlHandle_s* psHandle, long nCode, Job* pcJob )
{
	if( pcJob != NULL )
	{
		if( m_apcChildJobs.count( pcJob ) != 1 ) {
			DEBUG( "RecursiveDownloadJob: WARNING! Got TransferFinished() for job %x which isn't in our child list!\n", pcJob );
		}
		/* The child job should have cleaned itself up by now; remove it from our list. */
		m_apcChildJobs.erase( pcJob );
	}
	
	/* Iterate through the entry list and create & attach a Job for each entry. */
	/* m_pIterator should already be valid; it is initially set to m_acList.start() in HandleDirListData(). */
	if( m_pIterator == m_acList.end() )
	{
		/* Finished; clean up this job. Base class will take care of that, including notifying our parent (if any). */
		DirListJob::TransferFinished( psHandle, nCode, NULL );
		return;
	}

	Job* pcChild;
	String zDest = m_zLocalPath + "/" + m_pIterator->m_zName;
	int nNewID = m_pcServer->GetNewJobID();
	if( m_pIterator->IsDir() )
	{
		DEBUG( "RecursiveDownloadJob: Recursively getting dir '%s'  id %i\n", m_pIterator->m_zPath.c_str(), nNewID );
		pcChild = new RecursiveDownloadJob( m_pcServer, zDest, m_pIterator->m_zPath, nNewID, this );
	} else {
		DEBUG( "RecursiveDownloadJob: Recursively getting file '%s'  id %i\n", m_pIterator->m_zPath.c_str(), nNewID );
		pcChild = new DownloadJob( m_pcServer, zDest, m_pIterator->m_zPath, nNewID, this );
	}
	m_apcChildJobs.insert( pcChild );
	m_pcServer->AddJob( pcChild );
	pcChild->AttachToHandle( psHandle );
	/* When the child job finishes, it will call parent->TransferFinished() so that we end up back here. */
	
	m_pIterator++;
}

/**** RecursiveUploadJob implementation *****/
RecursiveUploadJob::RecursiveUploadJob( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nID, Job* pcParent )
	: Job( pcServer, zLocalPath, zRemotePath, JOB_RECURSIVE, nID, pcParent )
{
}

RecursiveUploadJob::~RecursiveUploadJob()
{
	if( m_pcDirIterator ) delete( m_pcDirIterator );
}


CURLcode RecursiveUploadJob::AttachToHandle( CurlHandle_s* psHandle )
{
	/* First open the local directory for listing; if it can't be opened then fail with error message */
	try {
		m_pcDirIterator = new Directory( m_zLocalPath );
	} catch( ... ) {
		printf( "RecursiveUploadJob: Error accessing directory '%s'!\n", m_zLocalPath.c_str() );
		return( CURLE_FAILED_INIT );
	}
	
	/* Then create & attach a child CommandJob to create the remote dir */
	String zCommand = "MKD ";
	zCommand += m_zRemotePath;
	std::vector<String> cTmp;
	cTmp.push_back( zCommand );
	
	int nNewID = m_pcServer->GetNewJobID();
	Job* pcChild = new CommandJob( m_pcServer, cTmp, nNewID, this );
	m_apcChildJobs.insert( pcChild );
	m_pcServer->AddJob( pcChild );
	pcChild->AttachToHandle( psHandle );
	
	/* The uploading will be done in TransferFinished() */

	return( CURLE_OK );
}

void RecursiveUploadJob::TransferFinished( CurlHandle_s* psHandle, long nCode, Job* pcJob )
{
	/* Detach previous job */
	if( pcJob != NULL )
	{
		if( m_apcChildJobs.count( pcJob ) != 1 ) {
			DEBUG( "RecursiveUploadJob: WARNING! Got TransferFinished() for job %x which isn't in our child list!\n", pcJob );
		}
		/* The child job should have cleaned itself up by now; remove it from our list. */
		m_apcChildJobs.erase( pcJob );
	}

	/* Get next dir entry, skipping . and .. */
	FileReference cFileRef;
	bool bDone = false;
	status_t nResult;
	while( !bDone ) {
		nResult = m_pcDirIterator->GetNextEntry( &cFileRef );
		/* As of Syllable 0.6.7, Directory::GetNextEntry() returns 1 on *success* and 0 on *end of directory* */
		if( nResult == 0 ) break;	/* Reached end of directory */
		if( cFileRef.IsValid() && cFileRef.GetName() != "." && cFileRef.GetName() != ".." ) bDone = true;
	}
	
	if(  nResult == 0 )
	{
//		DEBUG( "RecursiveUploadJob: reached end of directory (result %i, fileref %s)\n", nResult, cFileRef.IsValid()?cFileRef.GetName().c_str():"INVALID" );
		/* Reached end of directory. Clean up. */
		/* Base class will take care of cleanup, including notifying parent. */
		Job::TransferFinished( psHandle, nCode, NULL );
		return;
	}

	Job* pcChild;
	String zDest = m_zRemotePath + "/" + cFileRef.GetName();
	String zLocalPath;
	cFileRef.GetPath( &zLocalPath );
	struct stat sStat;
	cFileRef.GetStat( &sStat );	/* TODO: proper error handling */
	int nNewID = m_pcServer->GetNewJobID();
	if( S_ISDIR( sStat.st_mode ) )
	{
		DEBUG( "RecursiveUploadJob: Recursively sending dir '%s'  id %i\n", zLocalPath.c_str(), nNewID );
		pcChild = new RecursiveUploadJob( m_pcServer, zLocalPath, zDest, nNewID, this );
	} else {
		DEBUG( "RecursiveUploadJob: Recursively sending file '%s'  id %i\n", zLocalPath.c_str(), nNewID );
		pcChild = new UploadJob( m_pcServer, zLocalPath, zDest, nNewID, this );
	}
	m_apcChildJobs.insert( pcChild );
	m_pcServer->AddJob( pcChild );
	pcChild->AttachToHandle( psHandle );

}

bool RecursiveUploadJob::GetJobInfo( Message* pcMsg )
{
	return( false );
}

/**** RecursiveDeleteJob implementation ****/

RecursiveDeleteJob::RecursiveDeleteJob( Server* pcServer, const String& zRemotePath, int nID, Job* pcParent )
	: DirListJob( pcServer, zRemotePath, nID, pcParent )
{
	DEBUG( "Job %i is a RecursiveDeleteJob for '%s'\n", nID, zRemotePath.c_str() );
	m_nType = JOB_RECURSIVE;
	m_zRemotePath = zRemotePath;
	
	m_pIterator = m_acList.begin();	
	
	m_pcDeleteSelfJob = NULL;
}

RecursiveDeleteJob::~RecursiveDeleteJob()
{
}

CURLcode RecursiveDeleteJob::AttachToHandle( CurlHandle_s* psHandle )
{
	return( DirListJob::AttachToHandle( psHandle ) );
}

CURLcode RecursiveDeleteJob::RemoveFromHandle( CurlHandle_s* psHandle )
{
	return( DirListJob::RemoveFromHandle( psHandle ) );
}

void RecursiveDeleteJob::HandleDirListData( std::vector< RemoteNode >* pacList )
{
	m_acList.insert( m_acList.end(), pacList->begin(), pacList->end() );
	m_pIterator = m_acList.begin();
	delete( pacList );
}

void RecursiveDeleteJob::TransferFinished( CurlHandle_s* psHandle, long nCode, Job* pcJob )
{
	DEBUG( "RecursiveDeleteJob::TransferFinished( %li )\n", nCode );
	/* If a child job just finished, remove it from out list */
	if( pcJob != NULL )
	{
		if( m_apcChildJobs.count( pcJob ) != 1 ) {
			DEBUG( "RecursiveDeleteJob: WARNING! Got TransferFinished() for job %x which isn't in our child list!\n", pcJob );
		}
		/* The child job should have cleaned itself up by now; remove it from our list. */
		m_apcChildJobs.erase( pcJob );
	}
	
	/* If the final job, to rmdir this directory, has finished, then clean ourself up. */
	if( pcJob != NULL && pcJob == m_pcDeleteSelfJob )
	{
		/* Everything is finished. Clean up this job. Base class will take care of that, including notifying our parent (if any). */
		m_pcDeleteSelfJob = NULL;	/* It should have cleaned itself up by now */
		DirListJob::TransferFinished( psHandle, nCode, NULL );
		return;
	}
	
	if( m_pIterator == m_acList.end() )
	{
		/* Delete the current dir finally */
		int nNewID = m_pcServer->GetNewJobID();
		String zCommand = "RMD ";
		zCommand += m_zRemotePath;
		std::vector<String> cTmp;
		cTmp.push_back( zCommand );

		m_pcDeleteSelfJob = new CommandJob( m_pcServer, cTmp, nNewID, this );
		
		m_apcChildJobs.insert( m_pcDeleteSelfJob );
		m_pcServer->AddJob( m_pcDeleteSelfJob );
		m_pcDeleteSelfJob->AttachToHandle( psHandle );
		return;
	}

	/* Iterate through the entry list and create & attach a Job for each entry. */
	/* m_pIterator should already be valid; it is initially set to m_acList.start() in HandleDirListData(). */
	Job* pcChild;
	int nNewID = m_pcServer->GetNewJobID();
	if( m_pIterator->IsDir() )
	{
		DEBUG( "RecursiveDeleteJob: Recursively deleting dir '%s'  id %i\n", m_pIterator->m_zPath.c_str(), nNewID );
		pcChild = new RecursiveDeleteJob( m_pcServer, m_pIterator->m_zPath, nNewID, this );
	} else {
		DEBUG( "RecursiveDeleteJob: Recursively deleting file '%s'  id %i\n", m_pIterator->m_zPath.c_str(), nNewID );
		String zCommand = "DELE ";
		zCommand += m_pIterator->m_zPath;
		std::vector<String> cTmp;
		cTmp.push_back( zCommand );

		pcChild = new CommandJob( m_pcServer, cTmp, nNewID, this );
	}
	m_apcChildJobs.insert( pcChild );
	m_pcServer->AddJob( pcChild );
	pcChild->AttachToHandle( psHandle );
	/* When the child job finishes, it will call parent->TransferFinished() so that we end up back here. */
	
	m_pIterator++;	
}
