#ifndef __JOB_H__
#define __JOB_H__

#include <storage/file.h>
#include <util/string.h>
#include <util/handler.h>
#include <curl/curl.h>

#include <vector>
#include <set>

#include "remotenode.h"
#include "transferthread.h"

using namespace os;

// Prototype for the Server class.
class Server;
class os::Directory;

/* Job status enum */
enum {
	JOB_QUEUED = 1,
	JOB_RUNNING,
	JOB_PAUSED,
	JOB_DELETING
};

enum {
	JOB_DOWNLOAD,
	JOB_UPLOAD,
	JOB_DIRLIST,
	JOB_COMMAND,
	JOB_RECURSIVE
};

/** \brief Class representing a task, such as a single up/download or a recursive download.
 */
class Job
{
protected:
	Job( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nType, int nID, Job* pcParent = NULL );

public:
	virtual ~Job();	/* I want this to be private but can't be bothered debugging the compiler errors for now */
	
	/** Return the current status of the job.
	 * \todo Should this be here since it is a GUI access methods?
	 */
	int GetStatus()
	{
		return( m_nStatus );
	}
	
	/** \brief Return the type of the transaction.
	 * \todo Should this be here since it is a GUI access methods?
	 */
	int GetType()
	{
		return( m_nType );
	}
	
	/** \brief Determines whether the job should be automatically started when a curl handle becomes available.
	 * If this returns false, then TransferThread::_Schedule() will skip this job when choosing a job to
	 * attach to an available curl handle. If true, then it is eligible to be scheduled.
	 *
	 * Typically, 'top-level' jobs launched from the gui will return true.
	 * Child jobs created by a recursive job, eg RecursiveDownloadJob, will return false,
	 * as the parent job will explicitly decide when to run the child jobs.
	 */
	virtual bool IsSchedulable();
	
	/** \brief Return the current progress of the Job's operation.
	 * \todo Currently this just returns 0.
	 * \todo Should this be here since it is a GUI access methods?
	 */
	float GetProgress()
	{
		return( 0 );
	}
	
	String GetRemotePath()
	{
		return( m_zRemotePath );
	}
	
	String GetLocalPath()
	{
		return( m_zLocalPath );
	}

	/** \brief Find how much was downloaded so far (in bytes). */
	ssize_t GetDownloadedTotal()
	{
		return m_nTotalDownloaded;
	}
	
	/** \brief Find how much was downloaded so far (in bytes). */
	ssize_t GetUploadedTotal()
	{
		return m_nTotalUploaded;
	}

	/** \brief Find the actual size of the file being transferred (in bytes). */
	ssize_t GetTotalSize()
	{
		return m_nTotalSize;
	}
	
	/** \brief Save status information about the job into the given Message. */
	virtual bool GetJobInfo( Message* pcMsg );
	
private:

	/** \brief Read data into the provided buffer.
	 *
	 * \param pBuf Pointer to the buffer to read data into.
	 * \param nSize The amount of data to be read.
	 *
	 * \warning This is only a place holder and does nothing.
	 * It is meant to be subclassed by the different types of Job.
	 */
	virtual size_t Read( void* pBuf, size_t nSize )
	{
		printf( "Warning: Job::Read() called!\n" ); 
		return( 0 );
	}
	
	/** \brief Write data from the provided buffer.
	 *
	 * \param pBuf The location of the data that should be read from to write information.
	 * \param nSize Size of the data that is in pBuf.
	 * \warning This is only a place holder and does nothing.
	 * It is meant to be subclassed by the different types of Job.
	 */
	virtual size_t Write( void* pBuf, size_t nSize )
	{
		printf( "Warning: Job::Write() called!\n" );
		return( 0 );
	}

	/** \brief Seek into the data that is being manipulated.
	 * \param nOffset Amount to offset into the data being manipulated.
	 * \param nOrigin Where to start offsetting from.
	 * \warning This is only a place holder and does nothing.
	 * It is meant to be subclassed by the different types of Job.
	 */	
	virtual int Seek( curl_off_t nOffset, int nOrigin )
	{
		printf( "Warning: Job::Seek() called!\n" );
		return( -1 );
	}
	
	/** \brief Update the progress bar for the Job.
	 *
	 * \param fDownTotal Total data downloaded to be downloaded
	 * \param fDownNow Total data downloaded so far.
	 * \param fUpTotal Total data to be uploaded.
	 * \param fUpNow Total data uploaded so far.
	 *
	 * \warning This is only a place holder and does nothing.
	 * It is meant to be subclassed by the different types of Job.
	 */
	virtual int ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
	{
		printf( "Warning: Job::ProgressBar() called!\n" );
		return( 0 );
	}

	static size_t ReadCallback( void* pBuf, size_t nSize, size_t nMult, void* pCookie );
	static size_t WriteCallback( void* pBuf, size_t nSize, size_t nMult, void* pCookie );
	static int SeekCallback( void* pCookie, curl_off_t nOffset, int nOrigin );
	static int ProgressBarCallback( void* pCookie, double fDownTotal, double fDownNow, double fUpTotal, double fUpNow );

protected:
	/** \brief Set how much was downloaded so far.
	 * \param uSize The amount of data downloaded so far (in bytes).
	 */
	void SetDownloadedTotal(ssize_t nSize)
	{
		m_nTotalDownloaded = nSize;
	}

	/** \brief Set how much was uploaded so far.
	 * \param uSize The amount of data uploaded so far (in bytes).
	 */
	void SetUploadedTotal(ssize_t nSize)
	{
		m_nTotalUploaded = nSize;
	}
	
	/** \brief Set the actual size of the file being downloaded.
	 * \param uSize The size of the file (in bytes).
	 */
	void SetTotalSize(ssize_t nSize)
	{
		m_nTotalSize = nSize;
	}

public:	/* TODO - This should be protected but there's a compiler error which needs figuring out */
	virtual CURLcode AttachToHandle( CurlHandle_s* psHandle );
	virtual CURLcode RemoveFromHandle( CurlHandle_s* psHandle );
	
protected:
	virtual void Pause();
	virtual void Resume();
	virtual void Cancel();
	
	/** \brief Notification from curl that the data transfer has finished.
	 * This is called from the transfer thread when curl informs us that the data transfer for
	 * this job's curl handle is complete.
	 * For a simple standalone job, we would then want to release the curl handle, flush data to disk, clean up etc.
	 * For a recursive job, we might need to do something more complex.
	 * If the job is a child of a recursive job, we typically pass the call on to out parent by calling
	 * TransferFinished() on our parent with ourself as pcJob.
	 * \param psHandle The curl handle to which the job attached itself (via a previous AttachToHandle() call)
	 * \param nCode The result code from the server, eg 226 (transfer ok)
	 * \param pcJob If the call is originating from a child job, then pcJob is the originating child; otherwise NULL.
	 */
	 /* TODO: this should return 'true' if transfer thread should call _Schedule(), false if not. */
	virtual void TransferFinished( CurlHandle_s* psHandle, long nCode, Job* pcJob = NULL );

	/** \brief Server that this Job belongs to. */
	Server* m_pcServer;

	/** \brief Local path of the file to be transfered. */
	String m_zLocalPath;
	
	/** \brief Remote path of the file to be transfered. */
	String m_zRemotePath;
	
	/** \brief Stores type of transfer this job represents.
	 * Could be: Upload, Download, Directory Listing, etc.
	 *
	 * See the enums in the header file.
	 *
	 * \todo Should this be an enum type?
	 */
	int m_nType;
	
	/** \brief Stores the status of the job.
	 * Should use one of the enums for Active or Queued.
	 *
	 * See the enums in the header file.
	 *
	 * \todo Should this be an enum type?
	 */
	int m_nStatus;
	
	/** \brief Our id for the Server's std::map. */
	int m_nID;
	
	/** \brief Pointer to our parent Job (or NULL if none). This is used for child jobs of recursive jobs. */
	Job* m_pcParent;
	
	/** \brief The handle this job is currently attached to, or NULL if none. */
	CurlHandle_s* m_psHandle;
	
	/* Upload/download totals. */
	ssize_t m_nTotalSize;
	ssize_t m_nTotalDownloaded;
	ssize_t m_nTotalUploaded;
	
	friend class TransferThread;
	friend class Server;
};

/** \brief Single file download job.
 * This Job represents an individual file to be downloaded (as opposed to a directory).
 */
class DownloadJob : public Job
{
public:
	DownloadJob( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nID, Job* pcParent = NULL );
	~DownloadJob();

	bool GetJobInfo( Message* pcMsg );

protected:
	void Pause();
	void Cancel();
	
	void TransferFinished( CurlHandle_s* psHandle, long nCode, Job* pcJob = NULL );

	CURLcode AttachToHandle( CurlHandle_s* psHandle );
	CURLcode RemoveFromHandle( CurlHandle_s* psHandle );

	size_t Write( void* pBuf, size_t nSize );
	int Seek( curl_off_t nOffset, int nOrigin );
	int ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow );

private:
	File* m_pcFile; /**< A pointer to the file that is being downloaded. */
	
	friend class TransferThread;
};

/** \brief Single file upload job.
 * This Job represents an individual file to be uploaded.
 */
class UploadJob : public Job
{
public:

	UploadJob( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nID, Job* pcParent = NULL );
	~UploadJob();

	bool GetJobInfo( Message* pcMsg );

protected:
	CURLcode AttachToHandle( CurlHandle_s* psHandle );
	CURLcode RemoveFromHandle( CurlHandle_s* psHandle );
	
	size_t Read( void* pBuf, size_t nSize );
	int Seek( curl_off_t nOffset, int nOrigin );
	int ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow );

private:
	File* m_pcFile; /**< A pointer to the file that is being uploaded. */

	friend class TransferThread;
};

/** \brief Directory Listing Job.
 * This is a subclass of the Job type that is used for
 * directory listings.
 */
class DirListJob : public Job
{
public:
	DirListJob( Server* pcServer, const String& zRemotePath, int nID, Job* pcParent = NULL );
	~DirListJob();

	bool GetJobInfo( Message* pcMsg );
	
protected:
	CURLcode AttachToHandle( CurlHandle_s* psHandle );
	
	size_t Write( void* pBuf, size_t nSize );
	int ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow );

	size_t ParseDirListData( const char* pzBuf, size_t nSize, std::vector< RemoteNode >* pacList );
	
	virtual void HandleDirListData( std::vector< RemoteNode >* pacList );
	
	/** Whether or not this is the initial phase of the directory listing download. */
	bool m_bInitial;
	
	/** \brief Stores anything after last line ending with '\n'.
	 * This is used for when the read operation did read up to the new line.
	 * It is used to continue reading from the output of Curl. */
	String m_zLastLineFragment;

	friend class TransferThread;
};

/** \brief Directory listing initiated by the GUI
 *  This DirListNode sends the directory entries to the GUI.
 */
class GUIDirListJob : public DirListJob
{
public:
	GUIDirListJob( Server* pcServer, const String& zRemotePath, Handler* pcTarget, int nID, Job* pcParent = NULL );
	~GUIDirListJob();

protected:
	void TransferFinished( CurlHandle_s* psHandle, long nCode, Job* pcJob = NULL );
	
	void HandleDirListData( std::vector< RemoteNode >* pacList );

private:
	/** The target location to send the directory listing to. */
	Handler* m_pcTarget;
};

/** \brief A job that contains FTP commands we want to send to the server (eg 'delete').
 */
class CommandJob : public Job
{
public:
	CommandJob( Server* pcServer, const std::vector<String>& azCommands, int nID, Job* pcParent = NULL );
	~CommandJob();

	bool GetJobInfo( Message* pcMsg );

protected:
	CURLcode AttachToHandle( CurlHandle_s* psHandle );
	CURLcode RemoveFromHandle( CurlHandle_s* psHandle );
	
	std::vector<String> m_azCommands;
	
	struct curl_slist* m_psCommands;
	
	friend class TransferThread;
};


/** \brief A job representing a directory which we want to download recursively.
 *  It uses functionality from DirListJob but schedules new transfers rather than
 *  sending the data to the gui.
 */
class RecursiveDownloadJob : public DirListJob
{
public:
	RecursiveDownloadJob( Server* pcServer, const String& zLocalPath, const String& RemotePath, int nID, Job* pcParent = NULL );
	~RecursiveDownloadJob();
	
protected:
	CURLcode AttachToHandle( CurlHandle_s* psHandle );
	
	void HandleDirListData( std::vector< RemoteNode >* pacList );
	
	void TransferFinished( CurlHandle_s* psHandle, long nCode, Job* pcJob = NULL );

private:
	/** \brief List of the directory's contents */
	std::vector< RemoteNode > m_acList;
	
	/** \brief Iterator showing where we are up to while we download each directory list entry. */
	std::vector< RemoteNode >::iterator m_pIterator;
	
	/** \brief List of all our child jobs that currently exist */
	std::set< Job* > m_apcChildJobs;
	
	friend class TransferThread;
};

/** \brief A job representing a directory to be recursively uploaded to the server.
 */
class RecursiveUploadJob : public Job
{
public:
	RecursiveUploadJob( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nID, Job* pcParent = NULL );
	~RecursiveUploadJob();

	bool GetJobInfo( Message* pcMsg );

protected:
	CURLcode AttachToHandle( CurlHandle_s* psHandle );
	
	void TransferFinished( CurlHandle_s* psHandle, long nCode, Job* pcJob = NULL );

private:
	/** \brief Iterator for the local directory */
	Directory* m_pcDirIterator;

	/** \brief List of all our child jobs that currently exist. */
	std::set< Job* > m_apcChildJobs;
};


/** \brief A job representing a directory to be recursively deleted.
 *  It uses functionality from DirListJob but schedules new jobs rather than sending the data to the gui.
 */
class RecursiveDeleteJob : public DirListJob
{
public:
	RecursiveDeleteJob( Server* pcServer, const String& zRemotePath, int nID, Job* pcParent = NULL );
	~RecursiveDeleteJob();
	
protected:
	CURLcode AttachToHandle( CurlHandle_s* psHandle );
	CURLcode RemoveFromHandle( CurlHandle_s* psHandle );

	void HandleDirListData( std::vector< RemoteNode >* pacList );	
	
	void TransferFinished( CurlHandle_s* psHandle, long nCode, Job* pcJob = NULL );

private:
	/** \brief List of the directory's contents */
	std::vector< RemoteNode > m_acList;
	
	/** \brief Iterator showing where we are up to while we download each directory list entry. */
	std::vector< RemoteNode >::iterator m_pIterator;
	
	/** \brief List of all our child jobs that currently exist. */
	std::set< Job* > m_apcChildJobs;
	
	/** \brief The job that will rmdir this directory once all its contents are deleted. */
	Job* m_pcDeleteSelfJob;
	
	friend class TransferThread;
};


#endif	/* __JOB_H__ */

